/* style.c - Functions for creating and inserting styles.
 *
 * Copyright (C) 2001  Oskar Liljeblad
 *
 * This file is part of regex-markup.
 *
 * This software is copyrighted work licensed under the terms of the
 * GNU General Public License. Please consult the file `COPYING' for
 * details.
 */

#include <config.h>
#include "common/memory.h"
#include "common/strbuf.h"
#include "remark.h"

static void *
new_styleinfo(StyleInfoType type, const void *value)
{
	StyleInfo *info = xmalloc(sizeof(StyleInfo));
	info->type = type;
	info->value = value;
	return info;
}

StyleInfo *
new_styleinfo_style(Style *style)
{
	return new_styleinfo(STYLEINFO_STYLE, style);
}

StyleInfo *
new_styleinfo_pre(const char *string)
{
	return new_styleinfo(STYLEINFO_PRE, string);
}

StyleInfo *
new_styleinfo_post(const char *string)
{
	return new_styleinfo(STYLEINFO_POST, string);
}

Style *
new_style(const char *name, LList *spec)
{
	Style *style;
	Iterator *it;
	char *prebuf;
	char *postbuf;

	style = xmalloc(sizeof(Style));
	style->name = xstrdup(name);
	prebuf = strbuf_new();
	postbuf = strbuf_new();

	for (it = llist_iterator(spec); iterator_has_next(it); ) {
		StyleInfo *info = iterator_next(it);
		if (info->type == STYLEINFO_PRE) {
			strbuf_append(&prebuf, info->value);
		} else if (info->type == STYLEINFO_POST) {
			strbuf_prepend(&postbuf, info->value);
		} else if (info->type == STYLEINFO_STYLE) {
			const Style *add_style = info->value;
			strbuf_append(&prebuf, add_style->pre_string);
			strbuf_prepend(&postbuf, add_style->post_string);
		}
	}

	style->pre_string = strbuf_free_to_string(&prebuf);
	style->post_string = strbuf_free_to_string(&postbuf);
	style->refs = 1;

	return style;
}

void
free_style(Style *style)
{
	if (--style->refs <= 0) {
		free(style->name);
		free(style->pre_string);
		free(style->post_string);
		free(style);
	}
}

uint32_t
insert_style_pre(Style *style, char **strbuf, uint32_t pos)
{
	strbuf_insert(strbuf, pos, style->pre_string);
	return strlen(style->pre_string);
}

uint32_t
insert_style_post(Style *style, char **strbuf, uint32_t pos)
{
	strbuf_insert(strbuf, pos, style->post_string);
	return strlen(style->post_string);
}
